@@ -40,14 +40,18 @@
 #include "kchmsettings.h"
 #include "kchmsetupdialog.h"
 #include "iconstorage.h"
-#include "kchmviewwindow_qtextbrowser.h"
+#include "kchmnavtoolbar.h"
+#include "kchmviewwindow.h"
+#include "kchmviewwindowmgr.h"
+#include "kchmkeyeventfilter.h"
+#include "kchmcontentswindow.h"
 
-#if defined (USE_KDE)
-	#include "kchmviewwindow_khtmlpart.h"
-#else
+
+#if !defined (USE_KDE)
 	#include "kqrunprocess.h"
 #endif
 
+
 KCHMMainWindow::KCHMMainWindow()
     : KQMainWindow ( 0, "KCHMMainWindow", WDestructiveClose )
 {
@@ -59,9 +63,9 @@ KCHMMainWindow::KCHMMainWindow()
 	m_chmFile = 0;
 	
 	m_indexWindow = 0;
-	m_viewWindow = 0;
 	m_searchWindow = 0;
 	m_contentsWindow = 0;
+	m_viewWindowMgr = 0;
 
 	m_tabContextPage = -1;
 	m_tabIndexPage = -1;
@@ -74,14 +78,14 @@ KCHMMainWindow::KCHMMainWindow()
 		
 	// Create the initial layout - a splitter with tab window in left, and text browser in right
 	m_windowSplitter = new QSplitter(this);
-	m_tabWidget = new KQTabWidget (m_windowSplitter);
+	m_tabWidget = new KQTabWidget( m_windowSplitter );
+	m_viewWindowMgr = new KCHMViewWindowMgr( m_windowSplitter );
 	
 	m_bookmarkWindow = new KCHMBookmarkWindow (m_tabWidget);
 
 	// Add the tabs
-	m_tabWidget->addTab (m_bookmarkWindow, "Bookmarks");
+	m_tabWidget->addTab( m_bookmarkWindow, i18n("Bookmarks") );
 
-	createViewWindow();
 	setupToolbarsAndMenu();
 		
 	setCentralWidget( m_windowSplitter );
@@ -95,9 +99,11 @@ KCHMMainWindow::KCHMMainWindow()
 
 #if defined (ENABLE_AUTOTEST_SUPPORT)
 	m_autoteststate = STATE_OFF;
+	m_useShortAutotest = false;
 #endif /* defined (ENABLE_AUTOTEST_SUPPORT) */
 
 	QAccel * accel = new QAccel( this );
+	accel->connectItem ( accel->insertItem ( Key_F11 ), this, SLOT ( slotToggleFullScreenMode() ) );
 	accel->connectItem ( accel->insertItem ( CTRL + Key_1), this, SLOT ( slotActivateContentTab() ) );
 	accel->connectItem ( accel->insertItem ( CTRL + Key_2), this, SLOT ( slotActivateIndexTab() ) );
 	accel->connectItem ( accel->insertItem ( CTRL + Key_3), this, SLOT ( slotActivateSearchTab() ) );
@@ -105,6 +111,7 @@ KCHMMainWindow::KCHMMainWindow()
 	accel->connectItem ( accel->insertItem ( Key_F3 ), m_searchToolbar, SLOT ( onBtnNextSearchResult() ) );
 
 	statusBar()->show();
+	setIcon( *gIconStorage.getApplicationIcon() );
 }
 
 
@@ -116,9 +123,9 @@ KCHMMainWindow::~KCHMMainWindow()
 void KCHMMainWindow::slotOpenMenuItemActivated()
 {
 #if defined (USE_KDE)
-    QString fn = KFileDialog::getOpenFileName( appConfig.m_lastOpenedDir, "*.chm|Compressed Help Manual (*.chm)", this);
+    QString fn = KFileDialog::getOpenFileName( appConfig.m_lastOpenedDir, i18n("*.chm|Compressed Help Manual (*.chm)"), this);
 #else
-    QString fn = QFileDialog::getOpenFileName( appConfig.m_lastOpenedDir, "Compressed Help Manual (*.chm)", this);
+    QString fn = QFileDialog::getOpenFileName( appConfig.m_lastOpenedDir, i18n("Compressed Help Manual (*.chm)"), this);
 #endif
 
     if ( !fn.isEmpty() )
@@ -128,7 +135,7 @@ void KCHMMainWindow::slotOpenMenuItemActivated()
 		if ( !m_chmFile )
 			exit (1);
 			
-		statusBar()->message( tr("Loading aborted"), 2000 );
+		statusBar()->message( i18n("Loading aborted"), 2000 );
 	}
 }
 
@@ -147,6 +154,9 @@ bool KCHMMainWindow::loadChmFile ( const QString &fileName, bool call_open_page
 	
 		m_chmFile = new_chmfile;
 
+//		delete m_viewWindowMgr;
+		//m_viewWindowMgr = new KCHMViewWindowMgr( m_windowSplitter );
+		
 		QDir qd;
 		qd.setPath (fileName);
 		m_chmFilename = qd.absPath();
@@ -180,14 +190,15 @@ bool KCHMMainWindow::loadChmFile ( const QString &fileName, bool call_open_page
 		showOrHideSearchWindow( m_tabSearchPage );
 		
 		m_bookmarkWindow->invalidate();
-		m_viewWindow->invalidate();
-		updateView();
+		m_navToolbar->updateIconStatus( false, false );
+		m_viewWindowMgr->invalidate();
+		refreshCurrentBrowser();
 
 		if ( m_currentSettings->loadSettings (fileName) )
 		{
 			const KCHMTextEncoding::text_encoding_t * enc = KCHMTextEncoding::lookupByLCID (m_currentSettings->m_activeencodinglcid);
 
-			m_tabWidget->setCurrentPage (m_currentSettings->m_activetab);
+			m_tabWidget->setCurrentPage( m_currentSettings->m_activetabsystem );
 			
 			if ( enc )
 			{
@@ -201,10 +212,10 @@ bool KCHMMainWindow::loadChmFile ( const QString &fileName, bool call_open_page
 			m_bookmarkWindow->restoreSettings (m_currentSettings->m_bookmarks);
 
 			if ( call_open_page )
-				openPage (m_currentSettings->m_activepage, true);
-
-			m_viewWindow->setScrollbarPosition(m_currentSettings->m_scrollbarposition);
-			m_viewWindow->setZoomFactor(m_currentSettings->m_chosenzoom);
+			{
+				m_viewWindowMgr->restoreSettings( m_currentSettings->m_viewwindows );
+				m_viewWindowMgr->setCurrentPage( m_currentSettings->m_activetabwindow );
+			}
 		}
 		else
 		{
@@ -212,7 +223,7 @@ bool KCHMMainWindow::loadChmFile ( const QString &fileName, bool call_open_page
 			m_searchToolbar->setChosenEncodingInMenu (m_chmFile->getCurrentEncoding());
 			
 			if ( call_open_page )
-				openPage (m_chmFile->HomePage(), true);
+				openPage( m_chmFile->HomePage() );
 		}
 
 		m_searchToolbar->setEnabled (true);
@@ -224,12 +235,20 @@ bool KCHMMainWindow::loadChmFile ( const QString &fileName, bool call_open_page
 	{
 		if ( !m_chmFile )
 		{
-			QMessageBox mbox (tr("%1 - failed to load the chm file") . arg(APP_NAME), tr("Unable to load the chm file %1") . arg(fileName), QMessageBox::Critical, QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);
+			QMessageBox mbox(
+					i18n("%1 - failed to load the chm file") . arg(APP_NAME),
+					i18n("Unable to load the chm file %1") . arg(fileName), 
+					QMessageBox::Critical, 
+					QMessageBox::Ok, 
+					QMessageBox::NoButton, 
+					QMessageBox::NoButton);
 			mbox.exec();
 			exit (1);
 		}
 		
-		statusBar()->message( tr("Could not load file %1").arg(fileName), 2000 );
+		statusBar()->message( 
+				i18n("Could not load file %1").arg(fileName),
+				2000 );
 		delete new_chmfile;	
 		return false;
 	}
@@ -238,14 +257,19 @@ bool KCHMMainWindow::loadChmFile ( const QString &fileName, bool call_open_page
 
 void KCHMMainWindow::slotPrintMenuItemActivated()
 {
-	m_viewWindow->printCurrentPage();
+	getCurrentBrowser()->printCurrentPage();
 }
 
+
 void KCHMMainWindow::slotAboutMenuItemActivated()
 {
-	QString caption = tr("About %1") . arg(APP_NAME);
-	QString text = tr("%1 version %2\n\nCopyright (C) Georgy Yunaev, tim@krasnogorsk.ru, 2005\n\n"
-			"Licensed under GNU GPL license.") . arg(APP_NAME) . arg(APP_VERSION);
+	QString caption = i18n( "About %1" ) . arg(APP_NAME);
+	QString text = i18n( "%1 version %2\n\nCopyright (C) George Yunaev,"
+		"gyunaev@ulduzsoft.com, 2005-2006\nwww.kchmviewer.net\n\n"
+		"Licensed under GNU GPL license.\n\n"
+		"Please try our another project, www.transientmail.com - temporary "
+		"e-mail address, which expires automatically." )
+			. arg(APP_NAME) . arg(APP_VERSION);
 	
 	// It is quite funny that the argument order differs
 #if defined (USE_KDE)
@@ -261,17 +285,16 @@ void KCHMMainWindow::slotAboutQtMenuItemActivated()
     QMessageBox::aboutQt( this, APP_NAME);
 }
 
-void KCHMMainWindow::updateView( )
+void KCHMMainWindow::refreshCurrentBrowser( )
 {
 	QString title = m_chmFile->Title();
 	if ( !title )
 		title = APP_NAME;
 	else
 		title = (QString) APP_NAME + " - " + title;
-
-	setCaption (title);
+	setCaption ( title );
 	
-	m_viewWindow->invalidate();
+	getCurrentBrowser()->invalidate();
 	
 	if ( m_contentsWindow )
 	{
@@ -283,26 +306,39 @@ void KCHMMainWindow::updateView( )
 
 void KCHMMainWindow::slotOnTreeClicked( QListViewItem * item )
 {
+	bool unused;
+	
 	if ( !item )
 		return;
 	
 	KCHMMainTreeViewItem * treeitem = (KCHMMainTreeViewItem*) item;
-	
-	openPage(treeitem->getUrl(), false);
+	slotLinkClicked( treeitem->getUrl(), unused );
 }
 
 
 void KCHMMainWindow::slotLinkClicked ( const QString & link, bool& follow_link )
 {
-	// If the openPage failed, we do not need to follow the link.
-	follow_link = openPage( link );
+	if ( gKeyEventFilter.isShiftPressed() )
+	{
+		openPage( link, OPF_NEW_TAB | OPF_CONTENT_TREE );
+		follow_link = false;
+	}
+	else if ( gKeyEventFilter.isCtrlPressed() )
+	{
+		openPage( link, OPF_NEW_TAB | OPF_BACKGROUND );
+		follow_link = false;
+	}
+	else
+		// If the openPage failed, we do not need to follow the link.
+		follow_link = openPage( link, OPF_CONTENT_TREE | OPF_ADD2HISTORY );
 }
 
-bool KCHMMainWindow::openPage( const QString & srcurl, bool set_in_tree )
+
+bool KCHMMainWindow::openPage( const QString & srcurl, unsigned int flags )
 {
 	QString p1, p2, url = srcurl;
 
-	if ( m_viewWindow->isRemoteURL (url, p1) )
+	if ( getCurrentBrowser()->isRemoteURL (url, p1) )
 	{
 		switch ( appConfig.m_onExternalLinkClick )
 		{
@@ -311,12 +347,13 @@ bool KCHMMainWindow::openPage( const QString & srcurl, bool set_in_tree )
 
 		case KCHMConfig::ACTION_ASK_USER:
 	   		if ( QMessageBox::question(this,
-				tr ("%1 - remote link clicked - %2") . arg(APP_NAME) . arg(p1),
-           		tr ("A remote link %1 will start the external program to open it.\n\nDo you want to continue?").arg( url ),
-           		tr("&Yes"), tr("&No"),
-           		QString::null, 0, 1 ) )
-       				return false;
-				// no break!
+				 i18n("%1 - remote link clicked - %2") . arg(APP_NAME) . arg(p1),
+				 i18n("A remote link %1 will start the external program to open it.\n\nDo you want to continue?").arg( url ),
+				 i18n("&Yes"), i18n("&No"),
+				 QString::null, 0, 1 ) )
+					return false;
+				
+			// no break! should continue to open.
 
 		case KCHMConfig::ACTION_ALWAYS_OPEN:
 		{
@@ -333,27 +370,28 @@ bool KCHMMainWindow::openPage( const QString & srcurl, bool set_in_tree )
 	}
 		
 	// Filter the URLs which do not need to be opened at all by Qt version
-	if ( m_viewWindow->isJavascriptURL (url) )
+	if ( getCurrentBrowser()->isJavascriptURL (url) )
 	{
-		QMessageBox::information(this, tr ("%1 - JavsScript link clicked") . arg(APP_NAME),
-           	tr ("You have clicked a JavaScript link.\nBecause of security reasons, JavaScript links are disabled in CHM files."));
+		QMessageBox::information( this, 
+			i18n( "%1 - JavsScript link clicked") . arg(APP_NAME),
+			i18n( "You have clicked a JavaScript link.\nTo prevent security-related issues JavaScript URLs are disabled in CHM files.") );
 		
 		return false;
 	}
 
-	if ( m_viewWindow->isNewChmURL (url, p1, p2) 
+	if ( getCurrentBrowser()->isNewChmURL (url, p1, p2) 
 	&& p1 != m_chmFilename )
 	{
-   		if ( QMessageBox::question(this,
-			tr ("%1 - link to a new CHM file clicked") . arg(APP_NAME),
-           	tr ("You have clicked a link, which leads to a new CHM file %1.\nThe current file will be closed.\n\nDo you want to continue?").arg( p1 ),
-           	tr("&Yes"), tr("&No"),
-           	QString::null, 0, 1 ) )
-       			return false;
+   		if ( QMessageBox::question( this,
+			i18n( "%1 - link to a new CHM file clicked"). arg(APP_NAME),
+			i18n( "You have clicked a link, which leads to a new CHM file %1.\nThe current file will be closed.\n\nDo you want to continue?").arg( p1 ),
+			i18n( "&Yes" ), i18n( "&No" ),
+			QString::null, 0, 1 ) )
+				return false;
 
 		// Because chm file always contain relative link, and current filename is not changed,
 		// we need to form a new path
-		QFileInfo qfi (m_chmFilename);
+		QFileInfo qfi( m_chmFilename );
 
 		if ( !loadChmFile ( qfi.dirPath(true) + "/" + p1, false ) )
 			return false;
@@ -361,22 +399,22 @@ bool KCHMMainWindow::openPage( const QString & srcurl, bool set_in_tree )
 		url = p2;
 	}
 	
-	if ( m_viewWindow->openUrl (url) )
+	KCHMViewWindow * vwnd = getCurrentBrowser();
+	if ( flags & OPF_NEW_TAB )
+		vwnd = m_viewWindowMgr->addNewTab( !(flags & OPF_BACKGROUND) );
+	
+	// Store current page and position to add it to history if we change it
+	int hist_scrollpos = getCurrentBrowser()->getScrollbarPosition();
+	QString hist_url = getCurrentBrowser()->getOpenedPage();
+	
+	if ( vwnd->openUrl (url) )
 	{
 		// Open all the tree items to show current item (if needed)
-		KCHMMainTreeViewItem * treeitem;
-		if ( set_in_tree && (treeitem = m_chmFile->getTreeItem(m_viewWindow->getOpenedPage())) != 0 )
-		{
-			KCHMMainTreeViewItem * itemparent = treeitem;
-			while ( (itemparent = (KCHMMainTreeViewItem*) itemparent->parent()) != 0 )
-				itemparent->setOpen(true);
-			
-			if ( m_contentsWindow )
-			{
-				m_contentsWindow->setCurrentItem (treeitem);
-				m_contentsWindow->ensureItemVisible (treeitem);
-			}
-		}
+		if ( (flags & OPF_CONTENT_TREE) != 0 )
+			locateInContentTree( vwnd->getOpenedPage() );
+		
+		if ( flags & OPF_ADD2HISTORY )
+			getCurrentBrowser()->addNavigationHistory( hist_url, hist_scrollpos );
 	}
 	
 	return true;
@@ -407,122 +445,89 @@ void KCHMMainWindow::setupToolbarsAndMenu( )
 	// Create a 'file' toolbar
     QToolBar * toolbar = new QToolBar(this);
 	
-	toolbar->setLabel( tr("File Operations") );
+	toolbar->setLabel( i18n( "File Operations") );
 
     QPixmap iconFileOpen (*gIconStorage.getToolbarPixmap(KCHMIconStorage::fileopen));
     QToolButton * fileOpen = new QToolButton (iconFileOpen, 
-				tr("Open File"), 
+				i18n( "Open File" ), 
 				QString::null,
 				this, 
 				SLOT( slotOpenMenuItemActivated() ),
 				toolbar);
 	
-	QString fileOpenText = tr("Click this button to open an existing chm file.");
+	QString fileOpenText = i18n( "Click this button to open an existing chm file." );
 	QWhatsThis::add( fileOpen, fileOpenText );
 
     QPixmap iconFilePrint (*gIconStorage.getToolbarPixmap(KCHMIconStorage::print));
     QToolButton * filePrint	= new QToolButton (iconFilePrint,
-				tr("Print File"),
+				i18n( "Print File" ),
 				QString::null,
 				this,
 				SLOT( slotPrintMenuItemActivated() ),
 				toolbar);
 
-	QString filePrintText = tr("Click this button to print the current page");
+	QString filePrintText = i18n( "Click this button to print the current page");
 	QWhatsThis::add( filePrint, filePrintText );
 
-    QToolBar * navtoolbar = new QToolBar(this);
-	navtoolbar->setLabel( tr("Navigation") );
-	
-    QPixmap iconBackward (*gIconStorage.getToolbarPixmap(KCHMIconStorage::back));
-    m_toolbarIconBackward = new QToolButton (iconBackward,
-				tr("Move backward in history"),
-				QString::null,
-				this,
-				SLOT( slotBackwardMenuItemActivated() ),
-				navtoolbar);
-	QWhatsThis::add( m_toolbarIconBackward, tr("Click this button to move backward in browser history") );	
-
-    QPixmap iconForward (*gIconStorage.getToolbarPixmap(KCHMIconStorage::forward));
-    m_toolbarIconForward = new QToolButton (iconForward,
-				tr("Move forward in history"),
-				QString::null,
-				this,
-				SLOT( slotForwardMenuItemActivated() ),
-				navtoolbar);
-	QWhatsThis::add( m_toolbarIconBackward, tr("Click this button to move forward in browser history") );	
-	
-    QPixmap iconHome = (*gIconStorage.getToolbarPixmap(KCHMIconStorage::gohome));
-    new QToolButton (iconHome,
-				tr("Go to the home page"),
-				QString::null,
-				this,
-				SLOT( slotHomeMenuItemActivated() ),
-				navtoolbar);
-	QWhatsThis::add( m_toolbarIconBackward, tr("Click this button to move to the home page") );	
-
 	// Setup the menu
 	KQPopupMenu * file = new KQPopupMenu( this );
-	menuBar()->insertItem( tr("&File"), file );
+	menuBar()->insertItem( i18n( "&File"), file );
 
     int id;
-	id = file->insertItem ( iconFileOpen, tr("&Open..."), this, SLOT( slotOpenMenuItemActivated() ), CTRL+Key_O );
+	id = file->insertItem ( iconFileOpen, i18n( "&Open..."), this, SLOT( slotOpenMenuItemActivated() ), CTRL+Key_O );
     file->setWhatsThis( id, fileOpenText );
 
-	id = file->insertItem( iconFilePrint, tr("&Print..."), this, SLOT( slotPrintMenuItemActivated() ), CTRL+Key_P );
+	id = file->insertItem( iconFilePrint, i18n( "&Print..."), this, SLOT( slotPrintMenuItemActivated() ), CTRL+Key_P );
     file->setWhatsThis( id, filePrintText );
 
     file->insertSeparator();
-
+	id = file->insertItem ( i18n( "E&xtract CHM content..."), this, SLOT( slotExtractCHM() ) );
+	file->setWhatsThis( id, i18n( "Click this button to extract the whole CHM file content into a specific directory") );
+	file->insertSeparator();
+	
 	m_menuHistory = new KQPopupMenu( file );
 	connect ( m_menuHistory, SIGNAL( activated(int) ), this, SLOT ( slotHistoryMenuItemActivated(int) ));
 	
-	file->insertItem( tr("&Recent files"), m_menuHistory );
+	file->insertItem( i18n( "&Recent files"), m_menuHistory );
 	
 	file->insertSeparator();
-    file->insertItem( tr("&Quit"), qApp, SLOT( closeAllWindows() ), CTRL+Key_Q );
+	file->insertItem( i18n( "&Quit"), qApp, SLOT( closeAllWindows() ), CTRL+Key_Q );
 
 	KQPopupMenu * menu_edit = new KQPopupMenu( this );
-	menuBar()->insertItem( tr("&Edit"), menu_edit );
+	menuBar()->insertItem( i18n( "&Edit"), menu_edit );
 
-	id = menu_edit->insertItem ( tr("&Copy"), this, SLOT( slotBrowserCopy()), CTRL+Key_C );
-	id = menu_edit->insertItem ( tr("&Select all"), this, SLOT( slotBrowserSelectAll()), CTRL+Key_A );
+	menu_edit->insertItem ( i18n( "&Copy"), this, SLOT( slotBrowserCopy()), CTRL+Key_C );
+	menu_edit->insertItem ( i18n( "&Select all"), this, SLOT( slotBrowserSelectAll()), CTRL+Key_A );
 
     menu_edit->insertSeparator();
 	
+	// KCHMNavToolbar
+	m_navToolbar = new KCHMNavToolbar( this );
+	
 	// KCHMSearchToolbar also adds 'view' menu
 	m_searchToolbar = new KCHMSearchAndViewToolbar (this);
-
+	
+	// m_bookmarkWindow adds 'Bookmarks' menu
+	m_bookmarkWindow->createMenu( this );
+	
+	// m_viewWindowMgr creates 'Window' menu
+	m_viewWindowMgr->createMenu( this );
+	
 	KQPopupMenu * settings = new KQPopupMenu( this );
-	menuBar()->insertItem( tr("&Setup"), settings );
-	settings->insertItem( tr("&Change settings..."), this, SLOT( slotChangeSettingsMenuItemActivated() ));
+	menuBar()->insertItem( i18n( "&Settings"), settings );
+	settings->insertItem( i18n( "&Change settings..."), this, SLOT( slotChangeSettingsMenuItemActivated() ));
 
     KQPopupMenu * help = new KQPopupMenu( this );
-    menuBar()->insertItem( tr("&Help"), help );
+	menuBar()->insertItem( i18n( "&Help"), help );
 
-	help->insertItem( tr("&About"), this, SLOT( slotAboutMenuItemActivated() ), Key_F1 );
-	help->insertItem( tr("About &Qt"), this, SLOT( slotAboutQtMenuItemActivated() ));
+	help->insertItem( i18n( "&About"), this, SLOT( slotAboutMenuItemActivated() ), Key_F1 );
+	help->insertItem( i18n( "About &Qt"), this, SLOT( slotAboutQtMenuItemActivated() ));
     help->insertSeparator();
-    help->insertItem( tr("What's &This"), this, SLOT(whatsThis()), SHIFT+Key_F1 );
+	help->insertItem( i18n( "What's &This"), this, SLOT(whatsThis()), SHIFT+Key_F1 );
 	
 	updateHistoryMenu();
 }
 
-void KCHMMainWindow::slotBackwardMenuItemActivated()
-{
-	m_viewWindow->navigateBack();
-}
-
-void KCHMMainWindow::slotForwardMenuItemActivated()
-{
-	m_viewWindow->navigateForward();
-}
-
-void KCHMMainWindow::slotHomeMenuItemActivated()
-{
-	openPage (m_chmFile->HomePage(), true);
-}
-
 void KCHMMainWindow::slotAddBookmark( )
 {
 	emit m_bookmarkWindow->onAddBookmarkPressed ();
@@ -535,10 +540,12 @@ void KCHMMainWindow::setTextEncoding( const KCHMTextEncoding::text_encoding_t *
 	
 	// Because updateView() will call view->invalidate(), which clears the view->getOpenedPage(),
 	// we have to make a copy of it.
-	QString url = m_viewWindow->getOpenedPage();
-	updateView();
+	QString url = getCurrentBrowser()->getOpenedPage();
+	refreshCurrentBrowser();
 	
-	m_viewWindow->openUrl ( url );
+	// Regenerate the content and index trees
+	
+	getCurrentBrowser()->openUrl ( url );
 }
 
 void KCHMMainWindow::closeChmFile( )
@@ -547,16 +554,15 @@ void KCHMMainWindow::closeChmFile( )
 	if ( appConfig.m_HistoryStoreExtra )
 	{
 		m_currentSettings->m_activeencodinglcid = m_chmFile->getCurrentEncoding()->winlcid;
-		m_currentSettings->m_activetab = m_tabWidget->currentPageIndex( );
-		m_currentSettings->m_chosenzoom = m_viewWindow->getZoomFactor();
+		m_currentSettings->m_activetabsystem = m_tabWidget->currentPageIndex( );
+		m_currentSettings->m_activetabwindow = m_viewWindowMgr->currentPageIndex( );
 			
 		if ( m_searchWindow )
 			m_searchWindow->saveSettings (m_currentSettings->m_searchhistory);
 				
-		m_bookmarkWindow->saveSettings (m_currentSettings->m_bookmarks);
+		m_bookmarkWindow->saveSettings( m_currentSettings->m_bookmarks );
 
-		m_currentSettings->m_activepage = m_viewWindow->getOpenedPage();
-		m_currentSettings->m_scrollbarposition = m_viewWindow->getScrollbarPosition();
+		m_viewWindowMgr->saveSettings( m_currentSettings->m_viewwindows );
 
 		m_currentSettings->saveSettings( );
 	}
@@ -585,6 +591,9 @@ bool KCHMMainWindow::parseCmdLineArgs( )
 
 	if ( args->isSet("autotestmode") )
 		do_autotest = true;
+	
+	if ( args->isSet("shortautotestmode") )
+		do_autotest = m_useShortAutotest = true;
 /*
 	search_query = args->getOption ("search");
 	search_index = args->getOption ("sindex");
@@ -597,6 +606,8 @@ bool KCHMMainWindow::parseCmdLineArgs( )
 	for ( int i = 1; i < qApp->argc(); i++  )
 	{
 		if ( !strcmp (qApp->argv()[i], "--autotestmode") )
+			do_autotest = m_useShortAutotest = true;
+		else if ( !strcmp (qApp->argv()[i], "--shortautotestmode") )
 			do_autotest = true;
 		else if ( !strcmp (qApp->argv()[i], "--search") )
 			search_query = qApp->argv()[++i];
@@ -630,10 +641,11 @@ bool KCHMMainWindow::parseCmdLineArgs( )
 		if ( do_autotest )
 		{
 #if defined (ENABLE_AUTOTEST_SUPPORT)
-			if ( !filename.isEmpty() )
+			if ( filename.isEmpty() )
 				qFatal ("Could not use Auto Test mode without a chm file!");
 
 			m_autoteststate = STATE_INITIAL;
+			showMinimized ();
 			runAutoTest();
 #else
 			qFatal ("Auto Test mode support is not compiled in.");
@@ -645,39 +657,14 @@ bool KCHMMainWindow::parseCmdLineArgs( )
 	return false;
 }
 
-void KCHMMainWindow::slotHistoryAvailabilityChanged( bool enable_backward, bool enable_forward )
-{
-	m_toolbarIconBackward->setEnabled (enable_backward);
-	m_toolbarIconForward->setEnabled (enable_forward);
-}
-
-void KCHMMainWindow::createViewWindow( )
-{
-	if ( m_viewWindow )
-		delete m_viewWindow;
-
-#if defined (USE_KDE)
-	if ( !appConfig.m_kdeUseQTextBrowser )
-		m_viewWindow = new KCHMViewWindow_KHTMLPart ( m_windowSplitter );
-	else
-#endif
-		m_viewWindow = new KCHMViewWindow_QTextBrowser ( m_windowSplitter );
-	
-	// Handle clicking on link in browser window
-	connect( m_viewWindow->getQObject(), SIGNAL( signalLinkClicked (const QString &, bool &) ), this, SLOT( slotLinkClicked(const QString &, bool &) ) );
-
-	// Handle backward/forward buttons state change
-	connect( m_viewWindow->getQObject(), SIGNAL( signalHistoryAvailabilityChanged (bool, bool) ), this, SLOT( slotHistoryAvailabilityChanged (bool, bool) ) );
-}
-
 void KCHMMainWindow::slotBrowserSelectAll( )
 {
-	m_viewWindow->clipSelectAll();
+	getCurrentBrowser()->clipSelectAll();
 }
 
 void KCHMMainWindow::slotBrowserCopy( )
 {
-	m_viewWindow->clipCopy();
+	getCurrentBrowser()->clipCopy();
 }
 
 void KCHMMainWindow::slotChangeSettingsMenuItemActivated()
@@ -833,18 +820,15 @@ void KCHMMainWindow::showOrHideContextWindow( int tabindex )
 	{
 		if ( !m_contentsWindow )
 		{
-			m_contentsWindow = new KQListView (m_tabWidget);
-			m_contentsWindow->addColumn( "Contents" );
-			m_contentsWindow->setSorting(-1);
-			m_contentsWindow->setFocus();
-			m_contentsWindow->setRootIsDecorated(true);
-			m_contentsWindow->header()->hide();
-			m_contentsWindow->setShowToolTips(true);
-
-			// Handle clicking on m_contentsWindow element
-			connect( m_contentsWindow, SIGNAL( clicked( QListViewItem* ) ), this, SLOT( slotOnTreeClicked( QListViewItem* ) ) );
+			m_contentsWindow = new KCHMContentsWindow( m_tabWidget );
 			
-			m_tabWidget->insertTab (m_contentsWindow, tr("Contents"), tabindex);
+			// Handle clicking on m_contentsWindow element
+			connect( m_contentsWindow, 
+					 SIGNAL( clicked( QListViewItem* ) ), 
+					 this, 
+					 SLOT( slotOnTreeClicked( QListViewItem* ) ) );
+			
+			m_tabWidget->insertTab (m_contentsWindow, i18n( "Contents" ), tabindex);
 		}
 	}
 }
@@ -866,7 +850,7 @@ void KCHMMainWindow::showOrHideIndexWindow( int tabindex )
 		if ( !m_indexWindow )
 		{
 			m_indexWindow = new KCHMIndexWindow (m_tabWidget);
-			m_tabWidget->insertTab (m_indexWindow, tr("Index"), tabindex);
+			m_tabWidget->insertTab (m_indexWindow, i18n( "Index" ), tabindex);
 		}
 		else
 			m_indexWindow->invalidate();
@@ -889,7 +873,7 @@ void KCHMMainWindow::showOrHideSearchWindow( int tabindex )
 		if ( !m_searchWindow )
 		{
 			m_searchWindow = new KCHMSearchWindow (m_tabWidget);
-			m_tabWidget->insertTab (m_searchWindow, tr("Search"), tabindex);
+			m_tabWidget->insertTab (m_searchWindow, i18n( "Search" ), tabindex);
 		}
 		else
 			m_searchWindow->invalidate();
@@ -901,12 +885,20 @@ void KCHMMainWindow::slotEnableFullScreenMode( bool enable )
 	if ( enable )
 	{
 		if ( !isFullScreen() )
+		{
 			showFullScreen ();
+			menuBar()->hide();
+			statusBar()->hide();
+		}
 	}
 	else
 	{
 		if ( isFullScreen() )
+		{
 			showNormal ();
+			menuBar()->show();
+			statusBar()->show();
+		}
 	}
 }
 
@@ -918,6 +910,171 @@ void KCHMMainWindow::slotShowContentsWindow( bool show )
 		m_tabWidget->hide();
 }
 
+void KCHMMainWindow::slotToggleFullScreenMode( )
+{
+	slotEnableFullScreenMode( !isFullScreen() );
+}
+
+void KCHMMainWindow::slotLocateInContentWindow( )
+{
+	// There may be no content tab at all
+	if ( !m_contentsWindow  || m_tabContextPage == -1 )
+		return;
+	
+	// Activate a content tab
+	m_tabWidget->setCurrentPage( m_tabContextPage );
+	
+	// Open all the tree items to show current item (if needed)
+	KCHMMainTreeViewItem * treeitem = m_chmFile->getTreeItem( getCurrentBrowser()->getOpenedPage() );
+	if ( m_contentsWindow && treeitem )
+	{
+		KCHMMainTreeViewItem * itemparent = treeitem;
+		while ( (itemparent = (KCHMMainTreeViewItem*) itemparent->parent()) != 0 )
+			itemparent->setOpen(true);
+			
+		m_contentsWindow->setCurrentItem (treeitem);
+		m_contentsWindow->ensureItemVisible (treeitem);
+	}
+	else
+		statusBar()->message( i18n( "Could not locate opened topic in content window"), 2000 );
+}
+
+void KCHMMainWindow::slotExtractCHM( )
+{
+	QValueVector<QString> files;
+	
+#if defined (USE_KDE)
+	QString outdir = KFileDialog::getExistingDirectory (
+			QString::null,
+			this,
+			i18n("Choose a directory to store CHM content") );
+#else
+	QString outdir = QFileDialog::getExistingDirectory (
+			QString::null,
+			this,
+			0,
+			i18n("Choose a directory to store CHM content"),
+			TRUE );
+#endif
+	
+	if ( outdir.isEmpty() )
+		return;
+	
+	outdir += "/";
+	
+	// Enumerate all the files in archive
+	if ( !m_chmFile || !m_chmFile->enumerateArchive (files) )
+		return;
+
+	KQProgressModalDialog progress( i18n("Extracting CHM content"), i18n("Extracting files..."), i18n("Abort"), files.size(), this );
+	
+	for ( unsigned int i = 0; i < files.size(); i++ )
+	{
+		progress.setProgress( i );
+		
+		if ( (i % 3) == 0 )
+		{
+			qApp->processEvents();
+
+			if ( progress.wasCancelled() )
+				break;
+		}
+
+		// Extract the file
+		chmUnitInfo ui;	
+		if ( m_chmFile->ResolveObject( files[i], &ui ) )
+		{
+			QByteArray buf( ui.length );
+			
+			if ( m_chmFile->RetrieveObject (&ui, (unsigned char*) buf.data(), 0, ui.length) )
+			{
+				// Split filename to get the list of subdirectories
+				QStringList dirs = QStringList::split( '/', files[i] );
+
+				// Walk through the list of subdirectories, and create them if needed
+				// dirlevel is used to detect extra .. and prevent overwriting files
+				// outside the directory (like creating the file images/../../../../../etc/passwd
+				unsigned int i, dirlevel = 0;
+				QStringList dirlist;
+				
+				for ( i = 0; i < dirs.size() - 1; i++ )
+				{
+					// Skip .. which lead too far above
+					if ( dirs[i] == ".." )
+					{
+						if ( dirlevel > 0 )
+						{
+							dirlevel--;
+							dirlist.pop_back();
+						}
+					}
+					else
+					{
+						dirlist.push_back( dirs[i] );
+						
+						QDir dir ( outdir + dirlist.join( "/" ) );
+						if ( !dir.exists() )
+						{
+							if ( !dir.mkdir( dir.path() ) )
+								qWarning( "Could not create subdir %s\n", dir.path().ascii() );
+						}
+					}
+				}
+				
+				QString filename = outdir + dirlist.join( "/" ) + "/" + dirs[i];
+				QFile wf( filename );
+				if ( !wf.open( IO_WriteOnly ) )
+				{
+					 qWarning( "Could not write file %s\n", filename.ascii() );
+					 continue;
+				}
+				
+				wf. writeBlock( buf );
+				wf.close();
+			}
+		}
+		else
+			qWarning( "Could not resolve file %s\n", files[i].ascii() );
+	}
+	
+	progress.setProgress( files.size() );
+}
+
+KCHMViewWindow * KCHMMainWindow::getCurrentBrowser( ) const
+{
+	return m_viewWindowMgr->current();
+}
+
+void KCHMMainWindow::slotOpenPageInNewTab( )
+{
+	openPage( getCurrentBrowser()->getNewTabLink(), OPF_NEW_TAB | OPF_CONTENT_TREE );
+}
+
+void KCHMMainWindow::slotOpenPageInNewBackgroundTab( )
+{
+	openPage( getCurrentBrowser()->getNewTabLink(), OPF_NEW_TAB | OPF_BACKGROUND );
+}
+
+void KCHMMainWindow::slotBrowserChanged( KCHMViewWindow * newbrowser )
+{
+	locateInContentTree( newbrowser->getOpenedPage() );
+}
+
+void KCHMMainWindow::locateInContentTree( const QString & url )
+{
+	KCHMMainTreeViewItem * treeitem = m_chmFile->getTreeItem( url );
+	
+	if ( treeitem && m_contentsWindow )
+	{
+		KCHMMainTreeViewItem * itemparent = treeitem;
+		while ( (itemparent = (KCHMMainTreeViewItem*) itemparent->parent()) != 0 )
+			itemparent->setOpen(true);
+			
+		m_contentsWindow->setCurrentItem (treeitem);
+		m_contentsWindow->ensureItemVisible (treeitem);
+	}
+}
+
 
 #if defined (ENABLE_AUTOTEST_SUPPORT)
 void KCHMMainWindow::runAutoTest()
@@ -927,7 +1084,7 @@ void KCHMMainWindow::runAutoTest()
 	switch (m_autoteststate)
 	{
 	case STATE_INITIAL:
-		if ( m_contentsWindow )
+		if ( m_contentsWindow && !m_useShortAutotest )
 		{
 			m_autotestlistiterator = QListViewItemIterator (m_contentsWindow);
 			m_autoteststate = STATE_CONTENTS_OPENNEXTPAGE;
@@ -941,7 +1098,7 @@ void KCHMMainWindow::runAutoTest()
 	case STATE_CONTENTS_OPENNEXTPAGE:
 		if ( (item = (KCHMMainTreeViewItem *) m_autotestlistiterator.current()) != 0 )
 		{
-			openPage (item->getUrl(), true);
+			openPage( item->getUrl(), OPF_CONTENT_TREE | OPF_ADD2HISTORY );
 			m_autotestlistiterator++;
 		}
 		else
@@ -959,7 +1116,6 @@ void KCHMMainWindow::runAutoTest()
 		break;
 
 	case STATE_SHUTDOWN:
-		qDebug ("Autotest succeed");
 		qApp->quit();
 		break;
 		
@@ -968,3 +1124,6 @@ void KCHMMainWindow::runAutoTest()
 	}
 }
 #endif /* defined (ENABLE_AUTOTEST_SUPPORT) */
+
+
+#include "kchmmainwindow.moc"
