@@ -23,11 +23,7 @@
  *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
  ***************************************************************************/
 
-#include <qmessagebox.h> 
-#include <qstring.h>
-#include <qregexp.h>
-#include <qmap.h>
-#include <qeventloop.h>
+#include "kde-qt.h"
 
 #include "xchmfile.h"
 #include "kchmconfig.h"
@@ -48,6 +44,9 @@
 #include <qcursor.h>
 #include <qapplication.h>
 
+//#define DEBUGPARSER(A)	qDebug A
+#define DEBUGPARSER(A)	;
+
 class KCHMShowWaitCursor
 {
 public:
@@ -174,7 +173,9 @@ inline int CHMFile::findStringInQuotes (const QString& tag, int offset, QString&
 		m_entityDecodeMap["quot"] = "\""; // double quote
 		m_entityDecodeMap["apos"] = "'"; 	// single quote
 	}
+	
 	int qbegin = tag.find ('"', offset);
+	
 	if ( qbegin == -1 )
 		qFatal ("CHMFile::findStringInQuotes: cannot find first quote in <param> tag: '%s'", tag.ascii());
 
@@ -204,15 +205,33 @@ inline int CHMFile::findStringInQuotes (const QString& tag, int offset, QString&
 			{
 				if ( tag[i] == ';' ) // HTML entity ends
 				{
-					QMap<QString, QString>::const_iterator it = m_entityDecodeMap.find (htmlentity);
-					
-					if ( it == m_entityDecodeMap.end() )
+					// If entity is an ASCII code, just decode it
+					if ( htmlentity[0] == '#' )
 					{
-						qWarning ("CHMFile::DecodeHTMLUnicodeEntity: could not decode HTML entity '%s', abort decoding.", htmlentity.ascii());
-						break;
+						bool valid;
+						unsigned int ascode = htmlentity.mid(1).toUInt( &valid );
+						
+						if ( !valid )
+						{
+							qWarning ("CHMFile::DecodeHTMLUnicodeEntity: could not decode HTML entity '%s', abort decoding.", htmlentity.ascii());
+							break;
+						}
+
+						value.append ( QChar( ascode ) );
 					}
-	
-					value.append (it.data());
+					else
+					{
+						QMap<QString, QString>::const_iterator it = m_entityDecodeMap.find (htmlentity);
+					
+						if ( it == m_entityDecodeMap.end() )
+						{
+							qWarning ("CHMFile::DecodeHTMLUnicodeEntity: could not decode HTML entity '%s', abort decoding.", htmlentity.ascii());
+							break;
+						}
+						
+						value.append (it.data());
+					}
+
 					htmlentity = QString::null;
 					fill_entity = false;
 				}
@@ -333,11 +352,23 @@ bool CHMFile::ParseHhcAndFillTree (const QString& file, QListView *tree, bool as
 	if(src.isEmpty())
 		return false;
 
+#if 0
+	// Save the index for debugging purposes
+	QFile outfile( "parsed.htm" );
+	if ( outfile.open( IO_WriteOnly ) )
+	{
+		QTextStream textstream( &outfile );
+		textstream << src;
+		outfile.close();
+	}
+#endif
+	
 	unsigned int defaultimagenum = asIndex ? KCHMImageType::IMAGE_INDEX : KCHMImageType::IMAGE_AUTO;
 	unsigned int imagenum = defaultimagenum;
-	int pos = 0, indent = 0;
-	bool in_object = false, root_created = false;
+	int pos = 0, indent = 0, root_indent_offset = 0;
+	bool in_object = false, root_indent_offset_set = false;
 	bool add2treemap = asIndex ? false : m_treeUrlMap.isEmpty(); // do not add to the map during index search
+	
 	QString name;
 	QStringList urls;
 	KCHMMainTreeViewItem * rootentry[MAX_NEST_DEPTH];
@@ -394,34 +425,39 @@ bool CHMFile::ParseHhcAndFillTree (const QString& file, QListView *tree, bool as
 			{
 				KCHMMainTreeViewItem * item;
 
-				if ( !root_created )
-					indent = 0;
+				if ( !root_indent_offset_set )
+				{
+					root_indent_offset_set = true;
+					root_indent_offset = indent;
+					
+					if ( root_indent_offset > 1 )
+						qWarning("CHM has improper index; root indent offset is %d", root_indent_offset);
+				}
 
+				int real_indent = indent - root_indent_offset;
 				QString url = urls.join ("|");
 
-				// Add item into the tree
-				if ( !indent )
+				if ( real_indent == 0 )
 				{
-					item = new KCHMMainTreeViewItem (tree, lastchild[indent], name, url, imagenum);
+					// New root entry
+					item = new KCHMMainTreeViewItem (tree, lastchild[real_indent], name, url, imagenum);
+					DEBUGPARSER(("<root object>: '%s', new rootentry %08X\n", name.ascii(), item));
 				}
 				else
 				{
-					if ( !rootentry[indent-1] )
-						qFatal("CHMFile::ParseAndFillTopicsTree: child entry %d with no root entry!", indent-1);
+					// New non-root entry
+					if ( !rootentry[real_indent-1] )
+						qFatal("CHMFile::ParseAndFillTopicsTree: child entry \"%s\" indented as %d with no root entry!", name.ascii(), real_indent);
 
-					item = new KCHMMainTreeViewItem (rootentry[indent-1], lastchild[indent], name, url,  imagenum);
+					item = new KCHMMainTreeViewItem (rootentry[real_indent-1], lastchild[real_indent], name, url,  imagenum);
+					DEBUGPARSER(("<object>: '%s', indent %d, rootentry %08X, item %08X\n", name.ascii(), real_indent, rootentry[real_indent-1], item));
 				}
 
-				lastchild[indent] = item;
+				lastchild[real_indent] = item;
+				rootentry[real_indent] = item;
 
-				if ( indent == 0 || !rootentry[indent] )
-				{
-					rootentry[indent] = item;
-					root_created = true;
-
-					if ( asIndex  )
-						rootentry[indent]->setOpen(true);
-				}
+				if ( asIndex  )
+					rootentry[real_indent]->setOpen(true);
 
 				// There are no 'titles' in index file
 				if ( add2treemap  )
@@ -485,16 +521,30 @@ bool CHMFile::ParseHhcAndFillTree (const QString& file, QListView *tree, bool as
 			// Fix for buggy help files		
 			if ( ++indent >= MAX_NEST_DEPTH )
 				qFatal("CHMFile::ParseAndFillTopicsTree: max nest depth (%d) is reached, error in help file", MAX_NEST_DEPTH);
-				
+
 			lastchild[indent] = 0;
 			rootentry[indent] = 0;
+			
+			// This intended to fix <ul><ul>, which was seen in some buggy chm files,
+			// and brokes rootentry[indent-1] check
+			int real_indent = indent - root_indent_offset;
+			if ( real_indent > 1
+			&& rootentry[real_indent - 1] == 0
+			&& rootentry[real_indent - 2] != 0 )
+			{
+				rootentry[real_indent - 1] = rootentry[real_indent - 2];
+				qWarning("Broken CHM index/content: tree autocorrection enabled.");
+			}
+						  
+			DEBUGPARSER(("<ul>: new indent is %d, last rootentry was %08X\n", indent - root_indent_offset, rootentry[indent-1]));
 		}
 		else if ( tagword == "/ul" ) // decrease indent level
 		{
-			if ( --indent < 0 )
-				indent = 0;
-				
+			if ( --indent < root_indent_offset )
+				indent = root_indent_offset;
+
 			rootentry[indent] = 0;
+			DEBUGPARSER(("</ul>: new intent is %d\n", indent - root_indent_offset));
 		}
 
 		pos = i;	
@@ -506,6 +556,7 @@ bool CHMFile::ParseHhcAndFillTree (const QString& file, QListView *tree, bool as
 bool CHMFile::ParseAndFillTopicsTree(QListView *tree)
 {
 	KCHMShowWaitCursor wc;
+	m_treeUrlMap.clear();
 	return ParseHhcAndFillTree (m_topicsFile, tree, false);
 }
 
@@ -653,8 +704,13 @@ CHM_RESOLVE_SUCCESS;
 size_t CHMFile::RetrieveObject(const chmUnitInfo *ui, unsigned char *buffer,
 							   LONGUINT64 fileOffset, LONGINT64 bufferSize)
 {
-	return ::chm_retrieve_object(m_chmFile, const_cast<chmUnitInfo*>(ui), buffer, fileOffset,
-								 bufferSize);
+#if USE_BUILTIN_CHMLIB
+	return ::chm_retrieve_object(m_chmFile, ui, buffer, 
+		fileOffset, bufferSize);
+#else
+	return ::chm_retrieve_object(m_chmFile, const_cast<chmUnitInfo*>(ui),
+		buffer, fileOffset, bufferSize);
+#endif
 }
 
 
@@ -981,7 +1037,9 @@ inline bool CHMFile::InfoFromSystem()
  
 KCHMMainTreeViewItem * CHMFile::getTreeItem( const QString & str ) const
 {
-	KCHMTreeUrlMap_t::const_iterator it = m_treeUrlMap.find (str);
+	QString fixedstr = normalizePath( str );
+
+	KCHMTreeUrlMap_t::const_iterator it = m_treeUrlMap.find( fixedstr );
 	if ( it == m_treeUrlMap.end() )
 		return 0;
 		
@@ -1193,12 +1251,12 @@ QCString CHMFile::convertSearchWord( const QString & src )
 
 QString CHMFile::getTopicByUrl( const QString & search_url )
 {
+	unsigned char buf[COMMON_BUF_LEN];
+	
 	if ( !m_lookupTablesValid )
 		return QString::null;
 
-	unsigned char buf[COMMON_BUF_LEN];
-	int pos = search_url.find ('#');
-	QString fixedurl = pos == -1 ? search_url : search_url.left (pos);
+	QString fixedurl = normalizePath( search_url );
 
 	for ( unsigned int i = 0; i < m_chmTOPICS.length; i += TOPICS_ENTRY_LEN )
 	{
@@ -1277,166 +1335,11 @@ void CHMFile::GetSearchResults( const KCHMSearchProgressResults_t & tempres, KCH
 	}
 }
 
-/*
-bool CHMFile::ParseChmIndexFile ( const QString& file, bool asIndex, KCHMParsedIndexEntry_t & cont )
+
+QString CHMFile::normalizePath( const QString & path ) const
 {
-	QString src;
-	const int MAX_NEST_DEPTH = 256;
-
-	if ( file.isEmpty() || !GetFileContentAsString ( src, file ) )
-		return false;
-
-	unsigned int defaultimagenum = asIndex ? KCHMImageType::IMAGE_INDEX : KCHMImageType::IMAGE_AUTO;
-	unsigned int imagenum = defaultimagenum;
-	int pos = 0, indent = 0;
-	bool in_object = false, root_created = false;
-	QString name;
-	QStringList urls;
-	KCHMParsedIndexEntry * rootentry[MAX_NEST_DEPTH];
+	int pos = path.find ('#');
+	QString fixedpath = pos == -1 ? path : path.left (pos);
 	
-	memset (rootentry, 0, sizeof(*rootentry));
-	
-	// Split the HHC file by HTML tags
-	int stringlen = src.length();
-	
-	while ( pos < stringlen && (pos = src.find ('<', pos)) != -1 )
-	{
-		int i, word_end = 0;
-		
-		for ( i = ++pos; i < stringlen; i++ )
-		{
-			// If a " or ' is found, skip to the next one.
-			if ( (src[i] == '"' || src[i] == '\'') )
-			{
-				// find where quote ends, either by another quote, or by '>' symbol (some people don't know HTML)
-				int nextpos = src.find ( src[i], i+1 );
-				if ( nextpos == -1 	&& ( nextpos = src.find ('>', i+1) ) == -1 )
-				{
-					qWarning ("CHMFile::ParseHhcAndFillTree: corrupted TOC: %s", src.mid(i).ascii());
-					return false;
-				}
-
-				i =  nextpos;
-			}
-			else if ( src[i] == '>'  )
-				break;
-			else if ( !src[i].isLetterOrNumber() && src[i] != '/' && !word_end )
-				word_end = i;
-		}
-		
-		QString tagword, tag = src.mid (pos, i - pos);
-		 
-		if ( word_end )
-			tagword = src.mid (pos, word_end - pos).lower();
-		else
-			tagword = tag.lower();
-
-//qDebug ("tag: '%s', tagword: '%s'\n", tag.ascii(), tagword.ascii());
-						
-		// <OBJECT type="text/sitemap"> - a topic entry
-		if ( tagword == "object" && tag.find ("text/sitemap", 0, false) != -1 )
-			in_object = true;
-		else if ( tagword == "/object" && in_object ) 
-		{
-			// a topic entry closed. Add a tree item
-			if ( name )
-			{
-				KCHMParsedIndexEntry * item;
-
-				if ( !root_created )
-					indent = 0;
-
-				QString url = urls.join ("|");
-
-				// Add item into the tree
-				if ( !indent )
-				{
-					item = new KCHMParsedIndexEntry (name, 0, url, imagenum);
-				}
-				else
-				{
-					if ( !rootentry[indent-1] )
-						qFatal("CHMFile::ParseAndFillTopicsTree: child entry %d with no root entry!", indent-1);
-
-					item = new KCHMParsedIndexEntry (name, rootentry[indent-1], url, imagenum);
-				}
-
-				if ( indent == 0 || !rootentry[indent] )
-				{
-					rootentry[indent] = item;
-					root_created = true;
-				}
-			}
-			else
-			{
-				if ( !urls.isEmpty() )
-					qDebug ("CHMFile::ParseAndFillTopicsTree: <object> tag with url \"%s\" is parsed, but name is empty.", urls[0].ascii());
-				else
-					qDebug ("CHMFile::ParseAndFillTopicsTree: <object> tag is parsed, but both name and url are empty.");	
-			}
-
-			name = QString::null;
-			urls.clear();
-			in_object = false;
-			imagenum = defaultimagenum;
-		}
-		else if ( tagword == "param" && in_object )
-		{
-			// <param name="Name" value="First Page">
-			int offset; // strlen("param ")
-			QString name_pattern = "name=", value_pattern = "value=";
-			QString pname, pvalue;
-
-			if ( (offset = tag.find (name_pattern, 0, FALSE)) == -1 )
-				qFatal ("CHMFile::ParseAndFillTopicsTree: bad <param> tag '%s': no name=\n", tag.ascii());
-
-			// offset+5 skips 'name='
-			offset = findStringInQuotes (tag, offset + name_pattern.length(), pname, TRUE, FALSE);
-			pname = pname.lower();
-
-			if ( (offset = tag.find (value_pattern, offset, FALSE)) == -1 )
-				qFatal ("CHMFile::ParseAndFillTopicsTree: bad <param> tag '%s': no value=\n", tag.ascii());
-
-			// offset+6 skips 'value='
-			findStringInQuotes (tag, offset + value_pattern.length(), pvalue, FALSE, TRUE);
-
-//qDebug ("<param>: name '%s', value '%s'", pname.ascii(), pvalue.ascii());
-
-			if ( pname == "name" )
-				name = pvalue;
-			else if ( pname == "local" )
-				urls.push_back (KCHMViewWindow::makeURLabsoluteIfNeeded (pvalue));
-			else if ( pname == "see also" && asIndex && name != pvalue )
-				urls.push_back (":" + pvalue);
-			else if ( pname == "imagenumber" )
-			{
-				bool bok;
-				int imgnum = pvalue.toInt (&bok);
-	
-				if ( bok && imgnum >= 0 && (unsigned) imgnum < MAX_BUILTIN_ICONS )
-					imagenum = imgnum;
-			}
-		}
-		else if ( tagword == "ul" ) // increase indent level
-		{
-			// Fix for buggy help files		
-			if ( ++indent >= MAX_NEST_DEPTH )
-				qFatal("CHMFile::ParseAndFillTopicsTree: max nest depth (%d) is reached, error in help file", MAX_NEST_DEPTH);
-				
-			rootentry[indent] = 0;
-		}
-		else if ( tagword == "/ul" ) // decrease indent level
-		{
-			if ( --indent < 0 )
-				indent = 0;
-				
-			rootentry[indent] = 0;
-		}
-
-		pos = i;	
-	}
-	
-	return true;
-
+	return KCHMViewWindow::makeURLabsoluteIfNeeded( fixedpath );
 }
-*/
